/** This file is part of Reactor.
 *  Copyright (c) 2020-2024 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */
/* globals GridStack */

export const version = 24099;

import api from '/client/ClientAPI.js';
import Rulesets from "/client/Rulesets.js";
import Rule from "/client/Rule.js";
import Reaction from "/client/Reaction.js";
import Container from "/client/Container.js";
import Data from "/client/Data.js";

import Tab from "./tab.js";

import * as util from './reactor-ui-common.js';

import { _T, getLocale } from './i18n.js';

const reactionStatus = [
    _T( [ '#reaction-status-ended', "ended" ] ),
    _T( [ '#reaction-status-ready', "ready" ] ),
    _T( [ '#reaction-status-running', "running" ] ),
    _T( [ '#reaction-status-delayed', "delayed" ] ),
    _T( [ '#reaction-status-waiting', "waiting" ] ),
    _T( [ '#reaction-status-suspended', "suspended" ] )
];

const DEFAULT_LAYOUT = [
    { id: "widget0", w: 12, h: 2, content: '<div class="re-status-widget" data-widget="DefaultWidget"></div>' },
    { id: "widget1", w:  6, h: 2, content: '<div class="re-status-widget" data-widget="SetRules"></div>' },
    { id: "widget2", w:  6, h: 2, content: '<div class="re-status-widget" data-widget="RunningReactions"></div>' },
    { id: "widget3", w: 12, h: 2, content: '<div class="re-status-widget" data-widget="CurrentAlerts"></div>' }
];

const WIDGETS = [
    { id: "DefaultWidget", name: _T("New Widget") },
    { id: "SetRules", name: _T("Set Rules") },
    { id: "RunningReactions", name: _T("Running Reactions") },
    { id: "CurrentAlerts", name: _T("Current Alerts") },
    { id: "RecentEntities", name: _T("Recently Changed Entities") },
    { id: "RuleHistory", name: _T("Rule History") },
    { id: "ReactionHistory", name: _T("Reaction History") },
    { id: "ControllerStatus", name: _T("Controller Status") }
];

export default (function($) {

    const alertClasses = [ "bi-x-octagon-fill text-danger", "bi-exclamation-triangle-fill text-warning", "bi-info-circle-fill text-info" ];

    class StatusTab extends Tab {
        constructor( $parent ) {
            super( 'tab-status', $parent );
            this.lastCompact = 0;
            this.grid = false;
        }

        async activate( event ) {
            // console.log( this.$tab, "activate!", event );
            const $tab = this.$tab;
            const self = this;

            this.lastCompact = 0;
            this.grid = false;

            // $tab.addClass('w-100');

            const $stack = $( '<div class="grid-stack" gs-width="12"></div>' );

            function resizeGrid(grid) {
                const layout = "moveScale";
                let width = document.body.clientWidth;
                if (width < 700) {
                    grid.column(1, layout).cellHeight('100vw');
                } else if (width < 850) {
                    grid.column(3, layout).cellHeight('33.3333vw');
                } else if (width < 950) {
                    grid.column(6, layout).cellHeight('16.6667vw');
                } else if (width < 1100) {
                    grid.column(8, layout).cellHeight('12.25vw');
                } else {
                    grid.column(12, layout).cellHeight('8.3333vw');
                }
            }

            // Ref: https://www.jqueryscript.net/layout/Responsive-Fluid-Drag-and-Drop-Grid-Layout-with-jQuery-gridstack-js.html#jquery
            $stack.appendTo( $tab );
            this.grid = GridStack.init({
                    // auto: true,
                    // acceptWidgets: false,
                    animate: true,
                    column: 12,
                    // row: 6,
                    // height: 0,
                    margin: 8,
                    cellHeight: 133, /* was 20vw, fixed value for now ??? */
                    cellHeightUnit: 'px',
                    // alwaysShowResizeHandle: true,
                    disableResize: false,
                    disableOneColumnMode: true,
                    float: true,
                    removable: true,
                    removeTimeout: 2000
                });

            let layout = await this.loadGridLayout();
            this.grid.load( layout );

            /* "Add widget" widget */
            this.placeWidgetAdder( $stack );

            if ( false ) {
                $( 'div.re-status-widget', $stack ).each( ( ix, obj ) => {
                    const $c = $( obj ).closest( '.grid-stack-item' );
                    $( '<div class="re-status-widget-closer"><i class="bi bi-x"></i></div>' )
                        .hide()
                        .appendTo( $c );
                    $c.on( 'mouseenter', ( ev ) => {
                        // console.log( 'mouseenter', ev.shiftKey );
                        $( 'div.re-status-widget-closer', $( ev.currentTarget ) ).show();
                    }).on( 'mouseleave', ( ev ) => {
                        $( 'div.re-status-widget-closer', $( ev.currentTarget ) ).hide();
                    });
                });
            }

            resizeGrid(this.grid);
            $( window ).resize( () => {
                resizeGrid( self.grid );
            });

            $( 'div.re-status-widget', $tab ).each( ( ix, w ) => {
                const $w = $( w ).empty();
                const widgetName = $w.data( 'widget' );
                $w.addClass( 'card shadow' ).removeClass( (ix,name) =>
                    name.split(/\s+/).filter( el => el.match( /^border-/i ) ).join(' ') );
                try {
                    self[ `draw${widgetName}` ]( $w );
                } catch ( err ) {
                    $w.html( '<p>An error occurred while drawing the widget</p> ' );
                    $w.append( widgetName ).append( '; ' ).append( String( err ) );
                    console.log( err );
                }
            });

            const bound_save = this.saveGridLayout.bind( this );
            // this.grid.on( 'resizestop', bound_save );
            this.grid.on( 'change', bound_save );
            this.grid.on( 'removed', bound_save );
            this.grid.on( 'added', bound_save );

            $.ajax({
                dataType: 'json',
                url: '/api/v1/systime',
                timeout: 2000
            }).done( data => {
                let now = new Date();
                let hst = new Date( data.time );
                let diff = Math.abs( hst - now );
                console.log("UI TIME CHECK: Browser", now.toISOString(), "offset", -now.getTimezoneOffset(), "; host",
                    hst.toISOString(), "offset(mins)", data.tzoffset, "; diff(ms)", diff);
                if ( diff >= 5000 ) {
                    util.showSysModal({
                        title: _T('System Configuration Check'),
                        body: _T(['#sysconf-error-time',"The time on this system and on the Reactor host are significantly different. This may be due to incorrect system configuration on either or both. Please check the configuration of both systems. The host reports {0}; browser reports {1}; difference {2:.3f} seconds"],
                            hst.toISOString(), now.toISOString(), Math.floor( diff + 0.5 ) / 1000),
                        buttons: [{
                            class: "btn-primary",
                            label: "OK",
                            close: true
                        }]
                    });
                }
            }).fail( ( e ) => {
                console.log( "MSR host time check failed ", e );
            });
        }

        /** We don't implement suspend() because we have nothing to do. Let superclass do its work. */
        // suspend() {}

        placeWidgetAdder( $stack ) {
            if ( 0 === $( 'div.re-add-widget', $stack ).length ) {
                this.grid.addWidget( {id:'widget-adder',w:12,h:1,noResize:true,content:'<div class="re-add-widget"><i class="bi bi-plus-circle"></i></div>'} );
            }
            const self = this;
            $( 'div.re-add-widget', $stack )
                .addClass( 'align-self-center' )
                .attr( 'title', _T( 'Click to add a new widget' ) )
                .parent() /* to div.grid-stack-item-content */
                .addClass( 'd-flex flex-row justify-content-center' )
                .css( 'border', '1px solid #ccc')
                .off( 'click.reactor' )
                .on( 'click.reactor', async ( ev ) => {
                    const $me = $( ev.target ).closest( '.grid-stack-item' );
                    let wid = util.getUID( "widget" );
                    let widget = self.grid.addWidget({
                        id: wid,
                        x: 0,
                        y: parseInt( $me.attr( 'gs-y' ) ) || 0,
                        w: 6,
                        h: 2,
                        content: '<div class="re-status-widget card shadow border-danger" data-widget="DefaultWidget"></div>'
                    });
                    const $card = $( `div.re-status-widget`, $( widget ) );
                    await self.drawDefaultWidget( $card );
                    self.saveGridLayout();
                });
        }

        async drawDefaultWidget( $card ) {
            const self = this;
            $card.addClass( 'border-danger' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('About Status Panel Widgets')}</h6></div>` ).appendTo( $card );
            let $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $body.text( _T( '#status-widget-info-text' ) );
            $( `<div class="mt-2"> \
  <fieldset> \
    <label class="form-label">${_T( 'Choose Widget:' )}</label> \
    <select class="form-select re-status-widget-select"></select> \
  </fieldset> \
</div>` )
                .appendTo( $body );
            let $m = $( 'select.re-status-widget-select', $body );
            for ( let w of WIDGETS ) {
                $( '<option></option>' ).val( w.id ).text( w.name ).appendTo( $m );
            }
            $m.val( WIDGETS[0].id );
            $m.on( 'change', ( ev ) => {
                /* Widget type chosen. Set the type. */
                const sel = $( ev.target ).val();
                $card.data( 'widget', sel ).attr( 'data-widget', sel );

                /* Clear the default widget content and decorations and draw the new widget type */
                $card.empty()
                    .removeClass( (ix,name) => name.split(/\s+/).filter( el => el.match( /^border-/i ) ).join(' ') );
                self[`draw${sel}`]( $card );
            });
        }

        async drawSetRules( $card ) {
            $card.addClass( 'border-primary' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Set Rules')}</h6></div>` ).appendTo( $card );
            let $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $( '<div class="re-status-list"></div>' ).appendTo( $body );

            await this.updateSetRules( $card );
        }

        async updateSetRules( $card ) {
            const $list = $card.find( 'div.re-status-list' );
            var l = [];
            let rulesets = await Rulesets.getRulesets();
            await util.asyncForEach( rulesets, async set => {
                let rules = await set.getRules();
                await util.asyncForEach( rules, async rule => {
                    let dobj = await rule.getStatesDataObject();
                    let state = util.coalesce( ( ( dobj.value || {} ).rule || {} ).evalstate );
                    // console.log("updateSetRules set", set.name, "rule", rule.name, " states ",dobj.value);
                    if ( true === state ) {
                        l.push( { rule: rule, set: set, ts: dobj.value.rule.evalstamp } );
                    }

                    /* Subscribe to rule, so when it changes, we notice. */
                    this.subscribe( rule, this.updateSetRules.bind( this, $card ) );
                    this.subscribe( dobj, this.updateSetRules.bind( this, $card ) );
                });

                /* Subscribe to set, so when set or rule in set changes, list updates */
                this.subscribe( set, this.updateSetRules.bind( this, $card ) );
            });
            // console.log("updateSetRules",l.length,"set rules");
            $list.empty();
            l.sort( function( a, b ) {
                if ( a.ts === b.ts ) {
                    return a.rule.name.localeCompare( b.rule.name, undefined, { sensitivity: 'base' } );
                }
                return a.ts > b.ts ? -1 : 1;    /* N.B. reverse sort! */
            }).forEach( entry => {
                let $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' ).appendTo( $list );
                $( '<div class="col-5"></div>' ).text( entry.rule.name )
                    .toggleClass( 'text-danger', entry.rule.disabled )
                    .attr( 'title', entry.rule.disabled ? "This rule is disabled" : "" )
                    .appendTo( $row );
                $( '<div class="col-3"></div>' ).text( entry.set.name ).appendTo( $row );
                $( '<div class="col-4 text-end"></div>' ).text( util.shortTime( entry.ts ) ).appendTo( $row );
            });
            let $title = $( 'div.card-header h6', $list.closest( 'div.card' ) ).empty()
                .append( `<span>${ _T( 'Set Rules' ) }</span>` );
            if ( l.length > 0 ) {
                $title.append( $( '<span class="badge bg-primary ms-1">' ).text( l.length ) );
            } else {
                $list.text( _T( "There are no set rules at this moment." ) );
            }
        }

        async drawRunningReactions( $card ) {
            $card.addClass( 'border-info' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Running Reactions')}</h6></div>` ).appendTo( $card );
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            const $list = $( '<div class="re-status-list"></div>' ).appendTo( $body );
            $( '<div class="row no-gutters"><div class="col-auto text-align-center">Loading...</div></div>' )
                .appendTo( $list );

            let dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_queue' );
            if ( dobj ) {
                this.subscribe( dobj, this.updateRunningReactions.bind( this, $card ) );
            }

            await this.updateRunningReactions( $card );
        }

        insertCountDownTimer( $col, dt ) {
            function tick() {
                let delta = dt - Date.now();
                if ( delta >= 0 ) {
                    delta = Math.floor( ( delta + 999 ) / 1000 );
                    let ss = delta % 60;
                    ss = ss < 10 ? '0' + ss : ss;
                    let mm = Math.floor( delta / 60 );
                    mm = mm < 10 ? '0' + mm : mm;
                    $col.text( `${mm}:${ss}` );
                    setTimeout( () => { tick( $col, dt ); }, 500 );
                } else {
                    $col.empty();
                }
            }
            tick( $col, dt );
        }

        async updateRunningReactions( $card ) {
            const $list = $( 'div.card-body div.re-status-list', $card ).empty();
            let dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_queue' );
            let $title = $( 'div.card-header h6', $list.closest( 'div.card' ) );
            $( 'span.badge', $title ).remove();

            if ( Array.isArray( dobj.value ) && dobj.value.length > 0 ) {
                $title.append( $( '<span class="badge bg-success ms-1">' ).text( dobj.value.length ) );
                dobj.value.forEach( entry => {
                    let $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' )
                        .attr( 'id', 'running-' + entry.tid )
                        .attr( 'data-reaction', entry.id )
                        .data( 'reaction', entry.id )
                        .appendTo( $list );
                    // $( '<div class="col-auto"></div>' ).text( entry.tid ).appendTo( $row );
                    let $btn = $( '<div class="col-auto ps-1 pe-0"><button class="btn bi-btn bi-btn-sm btn-outline-danger"><i class="bi bi-x"></i></button></div>' )
                        .on( 'click', ev => {
                            const id = $( ev.currentTarget ).closest( 'div.row' ).data( 'reaction' );
                            console.log( 'stop reaction',id );
                            api.stopReaction( id, null, "user request" );
                        })
                        .appendTo( $row );
                    $( 'button', $btn ).attr( 'title', _T( 'Stop this reaction/task' ) );
                    $( '<div class="col ps-2 pe-0 re-task-name"></div>' ).appendTo( $row );
                    if ( entry.title ) {
                        $( '.re-task-name', $row ).text( entry.title + ' (' + entry.next_step + ')' );
                    } else if ( entry.rule ) {
                        /* Rule-based reaction. May be SET/RESET or subgroup */
                        Rule.getInstance( entry.rule ).then( rule => {
                            let reset = !!entry.id.match( /:R/ );
                            let nstep = ( ( ( reset ? rule.react_reset : rule.react_set ) || {}).actions || []).length;
                            $( '.re-task-name', $row ).text( rule.name + "<" + ( reset ? _T("RESET") : _T("SET") ) + '> (' +
                                entry.next_step + '/' + nstep + ')' );
                        });
                    } else {
                        Reaction.getInstance( entry.id ).then( reaction => {
                            let nstep = ( reaction.actions || [] ).length;
                            $( '.re-task-name', $row ).text( reaction.name + ' (' + entry.next_step + '/' + nstep + ')'  );
                        }).catch( err => {
                            console.error( err );
                            $col.text( String( err ) );
                        });
                    }
                    $( '<div class="col-auto ps-2 pe-0 text-center"></div>' ).text( reactionStatus[ entry.status + 1 ] || entry.status )
                        .appendTo( $row );
                    let $col = $( '<div class="col-auto ps-2 pe-1 text-end"></div>' ).appendTo( $row );
                    if ( 2 === entry.status && entry.time ) {
                        let delta = entry.time - Date.now();
                        if ( delta >= 3600000 ) {
                            $col.text( util.shortTime( entry.time ) );
                        } else {
                            this.insertCountDownTimer( $col, entry.time );
                        }
                    }
                });
            } else {
                $list.text( _T("There are no running reactions.") );
            }
        }

        async drawCurrentAlerts( $card ) {
            $card.addClass( 'border-success' );
            let $head = $( '<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"></div>' )
                .appendTo( $card );
            $( `<h6 class="m-0 fw-bold text-secondary">${_T('Current Alerts')}</h6>` )
                .appendTo( $head );
            $( `<div id="alertMenu" class="dropdown no-arrow">\
  <a class="dropdown-toggle" href="#" role="button" id="alertMenuLink" data-bs-toggle="dropdown" aria-haspopup="true" aria-expanded="false">\
    <i class="bi bi-three-dots-vertical text-gray-500"></i>\
  </a>\
  <div class="dropdown-menu dropdown-menu-end shadow" aria-labelledby="alertMenuLink" style="">\
    <!-- <div class="dropdown-header">Dropdown Header:</div> -->\
    <a id="re-alerts-clear" class="dropdown-item" href="#">${_T('Clear All')}</a>\
<!-- \
    <a class="dropdown-item" href="#">Another action</a>\
    <div class="dropdown-divider"></div>\
      <a class="dropdown-item" href="#">Something else here</a>\
    </div>\
-->\
  </div>\
</div>` ).appendTo( $head );
            $( 'a#re-alerts-clear', $head ).on( 'click.reactor', function() {
                api.clearAlerts();
            });
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $( '<div class="re-status-list">Loading alerts...</div>' ).appendTo( $body );

            let dobj = await Data.getInstance( Container.getInstance( 'states' ), 'sys_alerts' );
            this.subscribe( dobj, this.updateAlerts.bind( this, $card ) );

            await this.updateAlerts( $card );
        }

        async updateAlerts( $card ) {
            var $list = $( 'div.re-status-list', $card ).empty();

            let dobj = await Data.getInstance( Container.getInstance( 'states' ), 'sys_alerts' );

            let low = null;
            if ( 0 === dobj.value.alerts.length ) {
                $card.removeClass( "border-danger border-warning border-info" )
                    .addClass( "border-success" );
                $list.text( _T("There are no current alerts.") );
            } else {
                dobj.value.alerts.forEach( alert => {
                    if ( low === null || alert.severity < low ) {
                        low = alert.severity;
                    }
                    let $row = $( '<div class="row no-gutters py-1 align-items-start flex-nowrap"></div>' )
                        .attr( 'id', 'alert-' + alert.id )
                        .appendTo( $list );
                    $( '<div class="col-auto px-2"></div>' ).appendTo( $row )
                        .append( $( '<i class="bi bi-size-3"></i>' ).addClass( alertClasses[ alert.severity ] || "bi-x" ) );
                    let $col = $( '<div class="col pt-1"></div>' ).appendTo( $row );
                    $( '<div></div>' ).appendTo( $col ).text( alert.message );
                    if ( ( alert.options || {} ).detail ) {
                        let $l = $( '<div><tt></tt></div>' ).appendTo( $col );
                        $( 'tt', $l ).text( alert.options.detail.trim() );
                    }
                    $( '<div><small></small></div>' ).appendTo( $col );
                    $( 'small', $col ).text( _T( ['#alert-time-last', 'Last {0}'], util.shortTime( alert.ts ), alert.ts ) );
                    if ( alert.count && alert.count > 1 ) {
                        $( '<span></span>' ).text( _T("; this alert has repeated {0:d} times.", alert.count ) )
                            .appendTo( $( 'small', $col ) );
                    }
                    $( '<div class="col-auto pt-2 ps-2"></div>' ).appendTo( $row )
                        // .append( '<button class="btn bi-btn re-alert-go" title="Go To"><i class="bi bi-bullseye text-primary"></i></button>' )
                        .append( `<button class="btn bi-btn bi-btn-sm btn-outline-primary re-alert-ack" title="${_T('Dismiss Alert')}"><i class="bi bi-x text-primary"></i></button>` );
                });

                $card.removeClass( 'border-success' )
                    .toggleClass( 'border-danger', low === 0 )
                    .toggleClass( 'border-warning', low === 1 )
                    .toggleClass( 'border-info', low === 2 );
            }

            let $title = $( 'div.card-header h6', $list.closest( 'div.card' ) ).empty()
                .append( `<span>${_T('Current Alerts')}</span>` );
            if ( dobj.value.alerts.length > 0 ) {
                $title.append( $( '<span class="badge ms-1">' ).text( dobj.value.alerts.length )
                    .toggleClass( 'bg-danger', low === 0 )
                    .toggleClass( 'bg-warning', low === 1 )
                    .toggleClass( 'bg-info', low === 2 )
                );
            }

            $( 'button.re-alert-ack', $list ).on( 'click.reactor', (event) => {
                let $el = $( event.currentTarget ).closest( 'div.row' );
                let id = parseInt( $el.attr( 'id' ).replace( "alert-", "" ) );
                if ( ! isNaN( id ) ) {
                    api.dismissAlert( id );
                } else {
                    dobj.refresh();
                }
                return false;
            });
        }

        async drawRecentEntities( $card ) {
            const maxRows = 32;

            $card.addClass( 'border-secondary' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Recently Changed Entities')}</h6></div>` ).appendTo( $card );
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            const $list = $( '<div id="re-status-entities" class="re-status-list"></div>' )
                .attr( "data-maxrows", maxRows )
                .data( "maxrows", maxRows )
                .appendTo( $body );

            // Pre-populate the list.
            var l = api.getEntities();
            l.sort( (a, b) => {
                let t1 = a.getLastUpdate();
                let t2 = b.getLastUpdate();
                if ( t1 === t2 ) {
                    return a.getName().localeCompare( b.getName(), undefined, { sensitivity: 'base' } );
                }
                return t1 < t2 ? 1 : -1;    /* Reverse sort */
            });
            const numshow = parseInt( $list.data( 'maxrows' ) ) || 32;
            l.splice( numshow );
            l.forEach( e => {
                let $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' ).appendTo( $list );
                $( '<div class="col-xs-12 col-md-6 col-lg-4 overflow-hidden"></div>' ).text( e.getName() ).appendTo( $row );
                $( '<div class="col-xs-6 col-md-3 col-lg-4 overflow-hidden"></div>' ).text( e.getCanonicalID() ).appendTo( $row );
                $( '<div class="d-none d-lg-block col-lg-2 overflow-hidden"></div>' ).text( String( e.getPrimaryValue() ) ).appendTo( $row );
                $( '<div class="col-xs-6 col-md-3 col-lg-2 overflow-hidden text-end"></div>' ).text( util.shortTime( e.getLastUpdate() ) ).appendTo( $row );
            });

            api.off( "." + this.id );
            api.on( 'entity_change.' + this.id, this.updateRecentEntities.bind( this, $card ) );

            // ??? TO-DO: handle resize (changes maxrows)
        }

        async updateRecentEntities( $card, e ) {
            //console.log( 'status updateRecentEntities with entity', e );
            const $list = $( 'div#re-status-entities', $card );
            const $rows = $( 'div.row', $list );
            const numshow = parseInt( $list.data( 'maxrows' ) ) || 32;
            if ( $rows.length >= numshow ) {
                $rows[ numshow - 1 ].remove();
            }
            const $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' ).prependTo( $list );
            $( '<div class="col-xs-12 col-md-6 col-lg-4 overflow-hidden"></div>' ).text( e.getName() ).appendTo( $row );
            $( '<div class="col-xs-6 col-md-3 col-lg-4 overflow-hidden"></div>' ).text( e.getCanonicalID() ).appendTo( $row );
            $( '<div class="d-none d-lg-block col-lg-2 overflow-hidden"></div>' ).text( String( e.getPrimaryValue() ) ).appendTo( $row );
            $( '<div class="col-xs-6 col-md-3 col-lg-2 overflow-hidden text-end"></div>' ).text( util.shortTime( e.getLastUpdate() ) ).appendTo( $row );
        }

        async drawRuleHistory( $card ) {
            $card.addClass( 'border-primary' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Rule History')}</h6></div>` ).appendTo( $card );
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $( '<div class="re-status-list">Loading...</div>' ).appendTo( $body );

            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'rule_history' );
            this.subscribe( dobj, this.updateRuleHistory.bind( this, $card ) );

            await this.updateRuleHistory( $card );
        }

        async updateRuleHistory( $card ) {
            /* ??? Dynamic update, maybe 25 at a time until Y position > window length?
               Need to reaction to resize as well then.
            */
            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'rule_history' );
            const $list = $( 'div.re-status-list', $card ).empty();
            const n = dobj.value.history.length;
            for ( var ix = n-1; ix >= Math.max(0, n-100); ix-- ) {
                const e = dobj.value.history[ ix ];
				try {
					const rule = await Rule.getInstance( e.rule_id );
					const ruleset = await rule.getRuleset();
					let $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' ).appendTo( $list );
					$( '<div class="col-xs-12 col-md-7 overflow-hidden"></div>' )
						.text( `${rule.name} (${ruleset.name})` )
						.attr( 'title', e.rule_id )
						.appendTo( $row );
					$( '<div class="col-xs-2 col-md-2 overflow-hidden text-center"></div>' )
						.text( null === e.new_state ? _T("null") : ( e.new_state ? _T("SET") : _T("RESET") ) )
						.appendTo( $row );
					$( '<div class="col-xs-5 col-md-3 overflow-hidden text-end"></div>' )
						.text( util.shortTime( e.new_stamp ) )
						.appendTo( $row );
				} catch ( err ) {
					console.error( `Rule History Widget: can't display ${e.rule_id}: ${err.message}` );
				}
            }
        }

        async drawReactionHistory( $card ) {
            $card.addClass( 'border-info' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Reaction History')}</h6></div>` ).appendTo( $card );
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $( '<div class="re-status-list">Loading...</div>' ).appendTo( $body );

            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_history' );
            this.subscribe( dobj, this.updateReactionHistory.bind( this, $card ) );

            await this.updateReactionHistory( $card );
        }

        async updateReactionHistory( $card ) {
            /* ??? Dynamic update, maybe 25 at a time until Y position > window length?
               Need to reaction to resize as well then.
            */
            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_history' );
            const $list = $( 'div.re-status-list', $card ).empty();
            const n = dobj.value.history.length;
            for ( var ix = n-1; ix >= Math.max(0, n-100); ix-- ) {
                const e = dobj.value.history[ ix ];
                let $row = $( '<div class="row no-gutters py-1 align-items-start"></div>' ).appendTo( $list );
                $( '<div class="col-xs-6 col-md-6 overflow-hidden"></div>' )
                    .text( `${e.reaction_name}` )
                    .attr( 'title', e.reaction_id )
                    .appendTo( $row );
                $( '<div class="col-xs-3 col-md-3 overflow-hidden text-center"></div>' )
                    .text( util.shortTime( e.start_time ) )
                    .appendTo( $row );
                let st = true === e.result ? _T("OK") : String( e.result );
                st += `; ${(e.end_time||e.start_time)-e.start_time}ms`;
                $( '<div class="col-xs-3 col-md-3 overflow-hidden text-end"></div>' )
                    .text( st )
                    .appendTo( $row );
            }
        }

        async drawControllerStatus( $card ) {
            $card.addClass( 'border-info' );
            $( `<div class="card-header py-3 d-flex flex-row align-items-center justify-content-between"><h6 class="m-0 fw-bold text-secondary">${_T('Controller Status')}</h6></div>` ).appendTo( $card );
            const $body = $( '<div class="card-body pt-0 overflow-auto"></div>' ).appendTo( $card );
            $( '<div class="re-status-list">Loading...</div>' ).appendTo( $body );

            /* Subscribe for updates */
            const c = api.getControllers();
            const bound_update = this.updateControllerStatus.bind( this, $card );
            for ( let cid of c ) {
                let sys = api.getEntity( `${cid}>system` );
                if ( sys ) {
                    this.subscribe( sys, bound_update );
                }
            }

            this.updateControllerStatus( $card );
        }

        async updateControllerStatus( $card ) {
            const locale = getLocale();
            // console.log("updateControllerStatus for", $card, $card.closest('.grid-stack-item').attr('gs-id'));
            let c = api.getControllers();
            let l = [];
            for ( let cid of c ) {
                l.push( api.getEntity( `${cid}>system` ) );
            }
            const $list = $( 'div.re-status-list', $card );
            $list.empty();
            l.sort( ( a, b ) => {
                var aok = a.getAttribute( 'sys_system.state' ), bok = b.getAttribute( 'sys_system.state' );
                if ( aok === bok ) {
                    return a.getName().localeCompare( b.getName(), locale, { sensitivity: 'base' } );
                }
                return aok ? 1 : -1;
            });
            for ( let sys of l ) {
                const ctrl_id = sys.getCanonicalID().replace( />.*/, "" );
                const $row = $( '<div class="row"></div>' )
                    .attr( 'id', `ctrl-${ctrl_id}` )
                    .appendTo( $list );
                $( '<div class="col"></div>' ).text( `${sys.getName()}` )
                    .attr( 'title', _T(`Controller ID: {0}`, ctrl_id ) )
                    .appendTo( $row );
                let $col = $( '<div class="col-auto re-ctrl-status"></div>' ).appendTo( $row );
                if ( ! sys ) {
                    $col.html( '<i class="bi bi-question-square text-warning"></i>' );
                } else {
                    let state = sys.getAttribute( 'sys_system.state' );
                    if ( state ) {
                        $col.html( '<i class="bi bi-arrow-up-square text-success"></i>' );
                    } else if ( false === state ) {
                        $col.html( '<i class="bi bi-arrow-down-square-fill text-danger"></i>' );
                    } else {
                        $col.html( '<i class="bi bi-question-square text-warning"></i>' );
                    }
                }
if (false) {
                $col = $( '<div class="col-auto re-ctrl-status"></div>' ).appendTo( $row );
                if ( ! sys ) {
                    $col.html( '&nbsp;' );
                } else {
                    $( '<i class="bi bi-stop-circle text-danger re-ctrl-stop"></i>' )
                        .on( 'click', ( ev ) => {
                            const id = ( $( ev.currentTarget ).closest( 'div.row' ).attr( 'id' ) || "" ).replace( /^ctrl-/, "" );
                            api.stopController( id );
                        }).appendTo( $col );
                    $( '<i class="bi bi-play-circle text-success re-ctrl-start ms-1"></i>' )
                        .on( 'click', ( ev ) => {
                            const id = ( $( ev.currentTarget ).closest( 'div.row' ).attr( 'id' ) || "" ).replace( /^ctrl-/, "" );
                            api.startController( id );
                        }).appendTo( $col );
                    $( '<i class="bi bi-recycle text-warning re-ctrl-restart ms-1"></i>' )
                        .on( 'click', ( ev ) => {
                            const id = ( $( ev.currentTarget ).closest( 'div.row' ).attr( 'id' ) || "" ).replace( /^ctrl-/, "" );
                            api.restartController( id );
                        }).appendTo( $col );
                }
}
            }
        }

        async saveGridLayout( ev, el ) {
            // console.log( "saveGridLayout()",ev,el );

            /* Compress widget rows vertically only. Safety mechanism, we'll only do it once in a 500ms period.
             * We begin by iterating over the widgets and building an array (index=row) of arrays containing items
             * in that row (including rows of multi-row items). Then we iterate over that array and move items
             * up; any row with no items increases the gap; items are moved up by the gap count from their current
             * position.
             */
            if ( Date.now() > ( ( this.lastCompact || 0 ) + 500 ) ) {
                this.lastCompact = Date.now();
                let rows = [];
                let extent = 0;
                $( 'div.grid-stack-item', this.$tab ).each( (ix,w) => {
                    const $item = $( w );
                    const row = parseInt( $item.attr( 'gs-y' ) );
                    const h = parseInt( $item.attr( 'gs-h' ) );
                    for ( let n = 0; n < h; ++n ) {
                        rows[ row + n ] = rows[ row + n ] || [];
                        rows[ row + n ].push( { item: $item, n: n } );
                    }
                    if ( ( row + h - 1 ) > extent ) {
                        extent = row + h - 1;
                    }
                });
                let gap = 0;
                this.grid.batchUpdate();
                try {
                    for ( let row = 0; row <= extent; ++row ) {
                        if ( rows[ row ] && rows[ row ].length > 0 ) {
                            for ( let item of rows[ row ] || [] ) {
                                if ( 0 === item.n && item.item.attr( 'gs-y' ) != ( row - gap ) ) {
                                    this.grid.update( item.item.get(0), { y: row - gap } );
                                }
                            }
                        } else {
                            ++gap;
                        }
                    }
                } catch ( err ) {
                    console.err( err );
                } finally {
                    this.grid.commit();
                }
            }

            let layout = this.grid.save();
            /* Don't save the adder */
            let ix = layout.findIndex( el => el.id === "widget-adder" );
            if ( ix < 0 ) {
                /* Wasn't there at all! Restore it. */
                const $stack = $( 'div.grid-stack', this.$tab );
                setTimeout( () => { this.placeWidgetAdder( $stack ); }, 1 );
            }
            while ( ix >= 0 ) {
                layout.splice( ix, 1 );
                ix = layout.findIndex( el => el.id === "widget-adder" );
            }

            console.log( "Saving modified status layout to localStorage" );
            try {
                localStorage.setItem( 'reactor-status-layout', JSON.stringify( layout ) );
            } catch ( err ) {
                console.error( err );
            }
        }

        async loadGridLayout() {
            try {
                let layout = JSON.parse( localStorage.getItem( 'reactor-status-layout' ) );
                if ( Array.isArray( layout ) && layout.length > 0 ) {
                    console.log( "Using status widget layout loaded from localStorage" );
                    return layout;
                }
                console.error( "Invalid status widget layout; ignoring", layout );
            } catch ( err ) {
                console.error( err );
            }
            console.log( "Returning default status widget layout" );
            return DEFAULT_LAYOUT;
        }
    }

    var tabInstance = false;

    return {
        "init": function( $main ) {
            if ( ! tabInstance ) {
                tabInstance = new StatusTab( $main );
            }
        },
        "tab": () => tabInstance.getTabElement()
    };
})( jQuery );
